"""Query vlans of a device."""
from typing import Dict, Callable, List, Any

import pytest
from nornir.core.filter import F
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.filters import filter_hosts
from nuts.helpers.result import AbstractHostResultExtractor


class VlansExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[int, Any]:
        return self._cast_to_int(self._simple_extract(single_result)["vlans"])

    def _cast_to_int(self, elements: Dict[str, Any]) -> Dict[int, Any]:
        return {int(key): elements[key] for key in elements}


class VlansContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["vlans"]}

    def nornir_filter(self) -> F:
        return filter_hosts(self.nuts_parameters["test_data"])

    def nuts_extractor(self) -> VlansExtractor:
        return VlansExtractor(self)


CONTEXT = VlansContext


class TestNapalmVlans:
    @pytest.mark.nuts("vlan_tag")
    def test_vlan_tag(self, single_result, vlan_tag):
        assert vlan_tag in single_result.result

    @pytest.mark.nuts("vlan_name,vlan_tag")
    def test_vlan_name_to_tag(self, single_result, vlan_name, vlan_tag):
        assert single_result.result[vlan_tag]["name"] == vlan_name


class TestNapalmOnlyDefinedVlansExist:
    @pytest.mark.nuts("vlan_tags")
    def test_no_rogue_vlans(self, single_result, vlan_tags):
        assert list(single_result.result.keys()) == sorted(vlan_tags)


class TestNapalmInterfaceInVlan:
    @pytest.mark.nuts("vlan_tag,interface")
    def test_interface_port(self, single_result, vlan_tag, interface):
        assert interface in single_result.result[vlan_tag]["interfaces"]



"""Query config of a device."""
from typing import Dict, Callable, List, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.result import AbstractHostResultExtractor


class ConfigExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[int, Any]:
        return self._simple_extract(single_result)


class ConfigContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["config"]}

    def nuts_extractor(self) -> ConfigExtractor:
        return ConfigExtractor(self)


CONTEXT = ConfigContext


class TestNapalmConfig:
    @pytest.mark.nuts("startup_equals_running_config")
    def test_startup_equals_running_config(
        self, single_result, startup_equals_running_config
    ):
        assert (
            bool(
                single_result.result["config"]["startup"]
                == single_result.result["config"]["running"]
            )
            == startup_equals_running_config
        )



"""Query bandwidth performance between two devices."""
import pytest
import json
from typing import Dict, Callable, Any
import shlex

from nornir.core.task import Task, Result
from nornir.core.filter import F
from nornir_netmiko import netmiko_send_command

from nuts.context import NornirNutsContext
from nuts.helpers.errors import Error
from nuts.helpers.result import NutsResult, AbstractHostDestResultExtractor


class IperfExtractor(AbstractHostDestResultExtractor):
    def single_transform(self, single_result: Result) -> int:
        assert isinstance(single_result.result, str)
        return self._extract_bps(single_result.result)

    def _extract_bps(self, iperf_task_result: str) -> int:
        iperf_result = json.loads(iperf_task_result)
        if "error" in iperf_result:
            raise IperfResultError(iperf_result["error"])
        return int(iperf_result["end"]["sum_received"]["bits_per_second"])

    def single_result(self, nuts_test_entry: Dict[str, Any]) -> NutsResult:
        host = nuts_test_entry["host"]
        destination = nuts_test_entry["destination"]
        assert (
            host in self.transformed_result
        ), f"Host {host} not found in aggregated result."
        assert (
            destination in self.transformed_result[host]
        ), f"Destination {destination} not found in result."
        return self.transformed_result[host][destination]


class IperfContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return self.netmiko_run_iperf

    def netmiko_run_iperf(self, task: Task) -> Result:
        """
        Runs iperf between a host and several destinations.
        During setup, the destinations have been set up
        to act as servers.

        Note: The destination is not included in the nornir result
        if the iperf test fails. Therefore we cannot know which destination
        was not reachable, so we must patch the destination
        onto the result object to know later
        which host-destination pair actually failed.

        :param task: nornir task for iperf
        :return: All iperf results per host
        """
        destinations_per_host = [
            entry["destination"]
            for entry in self.nuts_parameters["test_data"]
            if entry["host"] == task.host.name
        ]
        for destination in destinations_per_host:
            escaped_dest = shlex.quote(destination)
            result = task.run(
                task=netmiko_send_command,
                command_string=f"iperf3 -c {escaped_dest} --json",
            )
            result[0].destination = destination  # type: ignore[attr-defined]
        return Result(host=task.host, result=f"iperf executed for {task.host}")

    def setup(self) -> None:
        """
        Sets up the all destinations to act as iperf servers.
        """
        test_data = self.nuts_parameters["test_data"]
        destinations = F(hostname__any={entry["destination"] for entry in test_data})
        assert self.nornir is not None
        selected_destinations = self.nornir.filter(destinations)
        selected_destinations.run(task=self.server_setup)

    def server_setup(self, task: Task) -> None:
        task.run(task=netmiko_send_command, command_string="iperf3 --server --daemon")

    def teardown(self) -> None:
        """
        Stops all destinations that acted as iperf servers.
        """
        test_data = self.nuts_parameters["test_data"]
        destinations = F(hostname__any={entry["destination"] for entry in test_data})
        assert self.nornir is not None
        selected_destinations = self.nornir.filter(destinations)
        selected_destinations.run(task=self.server_teardown)

    def server_teardown(self, task: Task) -> None:
        task.run(
            task=netmiko_send_command,
            command_string="pkill iperf3",
        )

    def nuts_extractor(self) -> IperfExtractor:
        return IperfExtractor(self)


CONTEXT = IperfContext


class IperfResultError(Error):

    """Error in iperf result JSON."""


class TestNetmikoIperf:
    @pytest.mark.nuts("min_expected")
    def test_iperf(self, single_result: NutsResult, min_expected: int) -> None:
        assert single_result.result >= min_expected



"""Allows to indicate only the test class name in a test bundle."""
from typing import Optional

default_index = {
    "TestNapalmBgpNeighbors": "nuts.base_tests.napalm_bgp_neighbors",
    "TestNapalmBgpNeighborsCount": "nuts.base_tests.napalm_bgp_neighbors",
    "TestNapalmInterfaces": "nuts.base_tests.napalm_interfaces",
    "TestNapalmLldpNeighbors": "nuts.base_tests.napalm_lldp_neighbors",
    "TestNapalmLldpNeighborsCount": "nuts.base_tests.napalm_lldp_neighbors",
    "TestNapalmNetworkInstances": "nuts.base_tests.napalm_network_instances",
    "TestNapalmPing": "nuts.base_tests.napalm_ping",
    "TestNapalmUsers": "nuts.base_tests.napalm_get_users",
    "TestNapalmConfig": "nuts.base_tests.napalm_get_config",
    "TestNapalmVlans": "nuts.base_tests.napalm_get_vlans",
    "TestNapalmOnlyDefinedVlansExist": "nuts.base_tests.napalm_get_vlans",
    "TestNapalmInterfaceInVlan": "nuts.base_tests.napalm_get_vlans",
    "TestNetmikoCdpNeighbors": "nuts.base_tests.netmiko_cdp_neighbors",
    "TestNetmikoCdpNeighborsCount": "nuts.base_tests.netmiko_cdp_neighbors",
    "TestNetmikoIperf": "nuts.base_tests.netmiko_iperf",
    "TestNetmikoOspfNeighborsCount": "nuts.base_tests.netmiko_ospf_neighbors",
    "TestNetmikoOspfNeighbors": "nuts.base_tests.netmiko_ospf_neighbors",
    "TestNapalmArp": "nuts.base_tests.napalm_get_arp",
    "TestNapalmArpRange": "nuts.base_tests.napalm_get_arp",
    "TestNapalmVrf": "nuts.base_tests.napalm_get_vrf",
}


def find_test_module_of_class(name: str) -> Optional[str]:
    """
    Try to resolve the name of the test class to a module.

    Test definitions can use an index to directly access a test
    by using the name of a test class instead of the module name.


    :param name: The name of the test class.
    :return: The path of the module which contains the test class.
    """
    return default_index.get(name, None)



"""Query BGP neighbors of a device or count them."""
from typing import Dict, Callable, List, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.result import AbstractHostResultExtractor, NutsResult


class BgpNeighborsExtractor(AbstractHostResultExtractor):
    def single_transform(
        self, single_result: MultiResult
    ) -> Dict[str, Dict[str, NutsResult]]:
        neighbors = self._simple_extract(single_result)["bgp_neighbors"]
        if "global" not in neighbors:
            return {}
        global_scope = neighbors["global"]
        router_id = global_scope["router_id"]
        return {
            peer: self._add_local_id(details, router_id)
            for peer, details in global_scope["peers"].items()
        }

    def _add_local_id(self, element: Dict[str, Any], router_id: str) -> Dict[str, Any]:
        element["local_id"] = router_id
        return element


class BgpNeighborsContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["bgp_neighbors"]}

    def nuts_extractor(self) -> BgpNeighborsExtractor:
        return BgpNeighborsExtractor(self)


CONTEXT = BgpNeighborsContext


class TestNapalmBgpNeighborsCount:
    @pytest.mark.nuts("neighbor_count")
    def test_neighbor_count(
        self, single_result: NutsResult, neighbor_count: int
    ) -> None:
        assert single_result.result is not None
        assert len(single_result.result) == neighbor_count


class TestNapalmBgpNeighbors:
    @pytest.fixture
    def peer_result(self, single_result: NutsResult, peer: str) -> Dict[str, Any]:
        assert single_result.result is not None
        return single_result.result[peer]

    @pytest.mark.nuts("peer,local_as")
    def test_local_as(self, peer_result: Dict[str, Any], local_as: int) -> None:
        assert peer_result["local_as"] == local_as

    @pytest.mark.nuts("peer,local_id")
    def test_local_id(self, peer_result: Dict[str, Any], local_id: str) -> None:
        assert peer_result["local_id"] == local_id

    @pytest.mark.nuts("peer,remote_as")
    def test_remote_as(self, peer_result: Dict[str, Any], remote_as: int) -> None:
        assert peer_result["remote_as"] == remote_as

    @pytest.mark.nuts("peer,remote_id")
    def test_remote_id(self, peer_result: Dict[str, Any], remote_id: str) -> None:
        assert peer_result["remote_id"] == remote_id

    @pytest.mark.nuts("peer,is_enabled")
    def test_is_enabled(self, peer_result: Dict[str, Any], is_enabled: bool) -> None:
        assert peer_result["is_enabled"] == is_enabled

    @pytest.mark.nuts("peer,is_up")
    def test_is_up(self, peer_result: Dict[str, Any], is_up: bool) -> None:
        assert peer_result["is_up"] == is_up



"""Query users of a device."""
from typing import Dict, Callable, List, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.result import AbstractHostResultExtractor, NutsResult


class UsersExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[str, Dict[str, Any]]:
        return self._simple_extract(single_result)["users"]


class UsersContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["users"]}

    def nuts_extractor(self) -> UsersExtractor:
        return UsersExtractor(self)


CONTEXT = UsersContext


class TestNapalmUsers:
    @pytest.mark.nuts("username")
    def test_username(self, single_result: NutsResult, username: str) -> None:
        assert username in single_result.result

    @pytest.mark.nuts("username,password")
    def test_password(
        self, single_result: NutsResult, username: str, password: str
    ) -> None:
        assert single_result.result[username]["password"] == password

    @pytest.mark.nuts("username,level")
    def test_privilege_level(
        self, single_result: NutsResult, username: str, level: int
    ) -> None:
        assert single_result.result[username]["level"] == level


class TestNapalmOnlyDefinedUsersExist:
    @pytest.mark.nuts("usernames")
    def test_no_rogue_users(
        self, single_result: NutsResult, usernames: List[str]
    ) -> None:
        assert list(single_result.result) == usernames



"""Fixtures"""
from typing import Optional, Dict, Any

import pytest
from _pytest.config.argparsing import Parser
from _pytest.main import Session
from _pytest.nodes import Collector
from _pytest.python import Metafunc
from _pytest.fixtures import FixtureRequest
from _pytest.config import Config
from py._path.local import LocalPath

from nuts.context import NutsContext
from nuts.context import NornirNutsContext
from nuts.helpers.result import NutsResult
from nuts.yamlloader import NutsYamlFile, get_parametrize_data


@pytest.fixture(scope="class")
def nuts_ctx(request: FixtureRequest) -> NutsContext:
    params = request.node.params
    context_class = getattr(request.module, "CONTEXT", NutsContext)
    ctx: NutsContext = context_class(params, pytestconfig=request.config)
    ctx.initialize()
    return ctx


@pytest.fixture
def single_result(nuts_ctx: NutsContext, nuts_test_entry: Dict[str, Any]) -> NutsResult:
    """
    Returns the result which belongs to a specific host
    out of the overall set of results that has been returned by nornir's task.
    In addition, ensures that the result has no exception and has not failed.

    :param nuts_ctx: The context for a test
    :param nuts_test_entry: The entry from the test bundle (yaml-file) for which
        the corresponding result should be returned
    :return: The `NutsResult` that belongs to a host or host/destination pair
    """
    res = nuts_ctx.extractor.single_result(nuts_test_entry)
    res.validate()
    return res


def pytest_configure(config: Config) -> None:
    config.addinivalue_line("markers", "nuts: marks the test for nuts parametrization")


def pytest_generate_tests(metafunc: Metafunc) -> None:
    """
    Checks if the the nuts pytest parametrization scheme exists (@pytest.mark.nuts)
    to generate tests based on that information.
    """
    nuts = metafunc.definition.get_closest_marker("nuts")
    if nuts:
        parametrize_args, parametrize_data = get_parametrize_data(
            metafunc, *nuts.args, **nuts.kwargs
        )
        metafunc.parametrize(parametrize_args, parametrize_data)


# https://docs.pytest.org/en/latest/example/nonpython.html#yaml-plugin
def pytest_collect_file(parent: Session, path: LocalPath) -> Optional[Collector]:
    """
    Performs the collection phase for the given pytest session.
    Collects all test bundles if available, i.e. files starting
    with 'test' and ending in .yaml.
    :param parent: pytest session object
    :param path: path to test file(s)
    :return: The pytest collector if found
    """
    if path.ext == ".yaml" and path.basename.startswith("test"):
        return NutsYamlFile.from_parent(parent, fspath=path)
    return None


def pytest_addoption(parser: Parser) -> None:
    """Add pytest command line options to configure nuts"""

    group = parser.getgroup("nuts")
    # Nornir context specific parameters
    group.addoption(
        "--nornir-config",
        "--nornir-configuration",
        action="store",
        dest="nornir_configuration",
        default=NornirNutsContext.DEFAULT_NORNIR_CONFIG_FILE,
        metavar="NORNIR_CONFIG",
        help="nuts nornir configuration file. Default is nr-config.yaml",
    )



"""Query network instances of a device."""
from typing import Dict, List, Callable, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.helpers.result import AbstractHostResultExtractor, NutsResult
from nuts.context import NornirNutsContext


class NetworkInstancesExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[str, Dict[str, Any]]:
        network_instances = self._simple_extract(single_result)["network_instances"]
        return {
            instance: self._transform_single_network_instance(details)
            for instance, details in network_instances.items()
        }

    def _transform_single_network_instance(
        self, network_instance: Dict[str, Any]
    ) -> Dict[str, Any]:
        return {
            "route_distinguisher": network_instance["state"]["route_distinguisher"],
            "interfaces": list(network_instance["interfaces"]["interface"]),
        }


class NetworkInstancesContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["network_instances"]}

    def nuts_extractor(self) -> NetworkInstancesExtractor:
        return NetworkInstancesExtractor(self)


CONTEXT = NetworkInstancesContext


class TestNapalmNetworkInstances:
    @pytest.mark.nuts("network_instance")
    def test_network_instance_exists(self, single_result, network_instance):
        assert network_instance in single_result.result

    @pytest.mark.nuts("network_instance,interfaces")
    def test_network_instance_contains_interfaces(
        self, single_result: NutsResult, network_instance: str, interfaces: List[str]
    ) -> None:
        assert single_result.result[network_instance]["interfaces"] == interfaces

    @pytest.mark.nuts("network_instance,route_distinguisher")
    def test_route_distinguisher(
        self, single_result: NutsResult, network_instance: str, route_distinguisher: str
    ) -> None:
        assert (
            single_result.result[network_instance]["route_distinguisher"]
            == route_distinguisher
        )



"""Results of a network query."""

import traceback
from typing import Any, Optional, TYPE_CHECKING, Dict

from nornir.core.task import MultiResult, AggregatedResult
from nuts.helpers.errors import NutsNornirError, NutsUnvalidatedResultError

if TYPE_CHECKING:
    from nuts.context import NutsContext


_TransformedResult = Dict[str, Any]


class NutsResult:
    """
    Potentially holds the result or context information if any issues occurred
    """

    def __init__(
        self,
        result: Optional[Any] = None,
        failed: Optional[bool] = False,
        exception: Optional[BaseException] = None,
    ) -> None:
        """
        Create a new NutsResult

        :param result: Result of the information gathering
        :param failed: Whether the information gathering failed or not
        :param exception: Exception which was thrown during the information gathering
        """
        self._result = result
        self.failed = failed
        self.exception = exception
        self._validated = False

    def validate(self) -> None:
        """Make sure the underlying result is a valid (i.e. non-failed) one."""
        if self.exception:
            header = "".join(
                traceback.format_exception_only(type(self.exception), self.exception)
            )
            raise NutsNornirError(
                f"An exception has occurred while executing nornir:\n"
                f"{header}\n"
                f"{self._result}"
            )
        if self.failed:
            raise NutsNornirError(f"Nornir execution has failed:\n" f"{self._result}")

        self._validated = True

    @property
    def result(self) -> Any:
        if not self._validated:
            raise NutsUnvalidatedResultError(
                f"Trying to access unvalidated result {self}"
            )
        return self._result


class AbstractResultExtractor:
    """Processes the general result that contains the raw overall results
    from the network query (mostly a nornir task)
    so that it can be passed on to the test class.

    Additionally wraps all the results that belong to one host into a
    NutsResult. There are two ways how this wrapping happens:

    1. Properties of a host are tested: One host maps to a NutsResult that contains
       all the queried properties.
    2. Properties of a connection between two hosts are tested: One host maps to a
       destination which maps to a NutsResult that contains all queried properties.
    """

    def __init__(self, context: "NutsContext") -> None:
        self._cached_result: Optional[_TransformedResult] = None
        self._nuts_ctx = context

    def single_transform(self, result: Any) -> Any:
        """
        Transforms a single (raw) result that belongs to a host from the overall
        general result set into the form that is required for the test class.
        :param result: raw data to be transformed
        :return: newly structured data
        """
        raise NotImplementedError

    def _map_host_to_nutsresult(
        self, general_result: AggregatedResult
    ) -> Dict[str, NutsResult]:
        """
        Maps a host's name to its corresponding result, which in turn is
        wrapped into a NutsResult.
        Used when a nornir tasks queries properties of a host.

        :param general_result: The raw result
                as provided by nornir's executed task
        :return: Host mapped to a NutsResult
        """
        return {
            host: self.nuts_result_wrapper(multiresult)
            for host, multiresult in general_result.items()
        }

    def _map_host_to_dest_to_nutsresult(
        self,
        general_result: AggregatedResult,
    ) -> Dict[str, Dict[str, NutsResult]]:
        """
        Maps a host's name to its corresponding destination and calls a helper function
        to further map that destination to a NutsResult.

        Used when a host-destination pair is tested.

        :param general_result: The raw result as provided by nornir's executed task
        :return: The host mapped to its corresponding destination
                  mapped to its NutsResult
        """
        return {
            host: self._map_dest_to_nutsresult(task_results)
            for host, task_results in general_result.items()
        }

    def _map_dest_to_nutsresult(
        self,
        task_results: MultiResult,
    ) -> Dict[str, NutsResult]:
        """
        Maps a destination to the result of a nornir task which both belong to a host.

        Note 1: The nornir Result object does not contain a destination property,
        which often means that if the specific task fails (e.g. a ping fails),
        the destination information is missing in the result, and we do not know
        which destination actually failed. It is therefore necessary to patch
        the destination onto the Result object to later know which host-destination pair
        actually failed. As a consequence, typing must ignore the patched attribute.

        Note 2: Why the first result is not used and task_results[1:] instead:
        Tasks per host (such as pinging a series of destinations) are
        nornir (sub)tasks that are executed from within another task.
        For the main task to finish properly, we must return a single Result, such as
        `Result(host=task.host, result="All pings executed")`, that confirms the
        fulfillment of the subtasks. All subsequent Result objects are then the
        completed subtasks. We do not need the first Result that simply confirms
        the fulfilled subtasks, therefore we skip it explicitly
        and return `task_results[1:]`.

        :param task_results: The results of the nornir task per host
        :return: The destination mapped to the desired information
        """
        return {
            single_result.destination: self.nuts_result_wrapper(  # type: ignore[attr-defined]  # noqa: E501
                single_result
            )
            for single_result in task_results[1:]
        }

    def nuts_result_wrapper(self, nornir_result: Any) -> NutsResult:
        """
        Wrap a nornir_result into a NutsResult

        :param nornir_result: The nornir_result which should be wrapped
        :return: NutsResult either containing the transformed result or
                 information if the network query has failed or thrown an exception.
        """
        if nornir_result.failed:
            return NutsResult(failed=True, exception=nornir_result.exception)
        try:

            return NutsResult(self.single_transform(nornir_result))
        except Exception as exception:
            tb = traceback.format_exc()
            return NutsResult(failed=True, exception=exception, result=tb)

    def transform_result(self, general_result: Any) -> _TransformedResult:
        """
        In most cases:
        1. calls map_host_to_nutsresult in case a host's properties are tested
        2. calls map_host_to_dest_to_nutsresult in case a host-destination pair's
           properties are tested.
        :param general_result: raw result
        :return: processed result ready to be passed to a test
        """
        raise NotImplementedError

    @property
    def transformed_result(
        self,
    ) -> _TransformedResult:
        """
        The (processed) results of the network task, ready to be passed on to a test's
        fixture. The results are cached, so that general_result does not need
        to be called multiple times as it might access the network.
        """
        if self._cached_result is None:
            self._cached_result = self.transform_result(self._nuts_ctx.general_result())
        return self._cached_result

    def single_result(self, nuts_test_entry: Dict[str, Any]) -> NutsResult:
        """The single result that belongs to one host.
        Must be overwritten in the case a host-destination pair is tested
        or the keyword that accesses nuts_test_entry differs.

        :param nuts_test_entry: a single entry in test_data
               of the test-bundle (the yaml-file)
        :return NutsResult of a host
        """
        host = nuts_test_entry["host"]
        assert (
            host in self.transformed_result
        ), f"Host {host} not found in aggregated result."
        return self.transformed_result[host]


class AbstractHostResultExtractor(AbstractResultExtractor):
    def transform_result(self, general_result: AggregatedResult) -> _TransformedResult:
        return self._map_host_to_nutsresult(general_result)

    def _simple_extract(self, single_result: MultiResult) -> Dict[Any, Any]:
        assert single_result[0].result is not None
        return single_result[0].result


class AbstractHostDestResultExtractor(AbstractResultExtractor):
    def transform_result(self, general_result: AggregatedResult) -> _TransformedResult:
        return self._map_host_to_dest_to_nutsresult(general_result)



"""Query CDP neighbors of a device."""
from typing import Callable, Dict, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_netmiko import netmiko_send_command

from nuts.helpers.result import AbstractHostResultExtractor, NutsResult
from nuts.context import NornirNutsContext


class CdpNeighborsExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[str, Dict[str, Any]]:
        neighbors = self._simple_extract(single_result)
        return {neighbor["destination_host"]: neighbor for neighbor in neighbors}


class CdpNeighborsContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return netmiko_send_command

    def nuts_arguments(self) -> Dict[str, Any]:
        return {"command_string": "show cdp neighbors detail", "use_textfsm": True}

    def nuts_extractor(self) -> CdpNeighborsExtractor:
        return CdpNeighborsExtractor(self)


CONTEXT = CdpNeighborsContext


class TestNetmikoCdpNeighbors:
    @pytest.mark.nuts("remote_host")
    def test_remote_host(self, single_result: NutsResult, remote_host: str) -> None:
        assert remote_host in single_result.result

    @pytest.mark.nuts("remote_host,local_port")
    def test_local_port(
        self, single_result: NutsResult, remote_host: str, local_port: str
    ) -> None:
        assert single_result.result[remote_host]["local_port"] == local_port

    @pytest.mark.nuts("remote_host,remote_port")
    def test_remote_port(
        self, single_result: NutsResult, remote_host: str, remote_port: str
    ) -> None:
        assert single_result.result[remote_host]["remote_port"] == remote_port

    @pytest.mark.nuts("remote_host,management_ip")
    def test_management_ip(
        self, single_result: NutsResult, remote_host: str, management_ip: str
    ) -> None:
        assert single_result.result[remote_host]["management_ip"] == management_ip


class TestNetmikoCdpNeighborsCount:
    @pytest.mark.nuts("neighbor_count")
    def test_neighbor_count(self, single_result, neighbor_count):
        assert neighbor_count == len(single_result.result)



class Error(Exception):
    """Base class for exceptions."""


class NutsSetupError(Error):
    """Errors caused during the setup of Nuts."""

    def __init__(self, message: str):
        super(NutsSetupError, self).__init__(message)


class NutsUsageError(Error):
    """Errors caused by the user."""


class NutsNornirError(Error):
    """Errors caused by nornir."""


class NutsUnvalidatedResultError(Error):
    """Internal error: An unvalidated NutsResult was accessed."""



"""
Functions to filter the nornir inventory and used in conjunction with
a context's nornir_filter function.
"""
from typing import Optional, Dict, Any, List, Set

from nornir.core.filter import F


def filter_hosts(test_data: Optional[List[Dict[str, Any]]]) -> F:
    assert test_data is not None
    hosts: Set[str] = {entry["host"] for entry in test_data}
    return F(name__any=hosts)



"""Query arp table of a device."""
from typing import Dict, Callable, List, Any, Text

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.result import AbstractHostResultExtractor


class ArpExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> List[Dict[Text, Any]]:
        result = []
        for entry in self._simple_extract(single_result)["arp_table"]:
            result.append({"interface": entry["interface"], "ip": entry["ip"]})
        return result


class ArpContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["arp_table"]}

    def nuts_extractor(self) -> ArpExtractor:
        return ArpExtractor(self)


CONTEXT = ArpContext


class TestNapalmArp:
    @pytest.mark.nuts("interface,ip")
    def test_arp_entry(self, single_result, interface, ip):
        assert single_result.result.count({"interface": interface, "ip": ip}) >= 1


class TestNapalmArpRange:
    @pytest.mark.nuts("min,max")
    def test_amount_of_arp_entries(self, single_result, min, max):
        assert min <= len(single_result.result) <= max



"""Query OSPF neighbors of a device or count them."""
from typing import Callable, Dict, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_netmiko import netmiko_send_command

from nuts.context import NornirNutsContext
from nuts.helpers.result import AbstractHostResultExtractor, NutsResult


class OspfNeighborsExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[str, Any]:
        neighbors = self._simple_extract(single_result)
        return {details["neighbor_id"]: details for details in neighbors}


class OspfNeighborsContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return netmiko_send_command

    def nuts_arguments(self) -> Dict[str, Any]:
        return {"command_string": "show ip ospf neighbor", "use_textfsm": True}

    def nuts_extractor(self) -> OspfNeighborsExtractor:
        return OspfNeighborsExtractor(self)


CONTEXT = OspfNeighborsContext


class TestNetmikoOspfNeighborsCount:
    @pytest.mark.nuts("neighbor_count")
    def test_neighbor_count(
        self, single_result: NutsResult, neighbor_count: int
    ) -> None:
        assert len(single_result.result) == neighbor_count


class TestNetmikoOspfNeighbors:
    @pytest.mark.nuts("neighbor_id")
    def test_neighbor_id(self, single_result: NutsResult, neighbor_id: str) -> None:
        assert neighbor_id in single_result.result

    @pytest.mark.nuts("neighbor_id,neighbor_address")
    def test_neighbor_address(
        self, single_result: NutsResult, neighbor_id: str, neighbor_address: str
    ) -> None:
        neighbor = single_result.result[neighbor_id]
        assert neighbor["address"] == neighbor_address

    @pytest.mark.nuts("local_port,neighbor_id")
    def test_local_port(
        self, single_result: NutsResult, local_port: str, neighbor_id: str
    ) -> None:
        neighbor = single_result.result[neighbor_id]
        assert neighbor["interface"] == local_port

    @pytest.mark.nuts("neighbor_id,state")
    def test_state(
        self, single_result: NutsResult, neighbor_id: str, state: str
    ) -> None:
        neighbor = single_result.result[neighbor_id]
        assert neighbor["state"] == state



"""Provide necessary information that is needed for a specific test."""
import pathlib
from typing import Any, Callable, Optional, Dict
from _pytest.config import Config

from nornir import InitNornir
from nornir.core import Nornir
from nornir.core.task import AggregatedResult, Result
from nornir.core.filter import F

from nuts.helpers.errors import NutsSetupError
from nuts.helpers.result import AbstractResultExtractor
from nuts.helpers.filters import filter_hosts


class NutsContext:
    """
    Base context class. Holds all necessary information that is needed
    for a specific test.

    :param nuts_parameters: test-specific data that is defined in the test bundle,
        i.e. the yaml file that is converted to nuts tests
    """

    def __init__(
        self, nuts_parameters: Any = None, pytestconfig: Optional[Config] = None
    ):
        self.nuts_parameters = nuts_parameters or {}
        self.extractor = self.nuts_extractor()
        self._pytestconfig = pytestconfig

    def initialize(self) -> None:
        """Initialize dependencies for this context after it has been created."""
        pass

    def nuts_extractor(self) -> AbstractResultExtractor:
        """Get a result extractor for this context."""
        return AbstractResultExtractor(self)

    def nuts_arguments(self) -> Dict[str, Any]:
        """
        Additional arguments for the (network) task to be executed.
        These can also be parameters that are defined in the `test_execution`
        part of the test bundle.

        If the subclass is a NornirNutsContext, the arguments are passed on
        to the nornir task:
        Note that the arguments then must match those that the nornir task offers.

        :return: A dict containing the additional arguments
        """
        test_execution = self.nuts_parameters.get("test_execution", None)
        return {**(test_execution if test_execution is not None else {})}

    def general_result(self) -> Any:
        """
        :return: raw, unprocessed result
        """
        raise NotImplementedError

    @property
    def pytestconfig(self) -> Optional[Config]:
        """
        Set the pytest configuration.

        Can be overwritten to aggregate the configuration
        """
        return self._pytestconfig


class NornirNutsContext(NutsContext):
    """
    NutsContext class which provides nornir-specific helpers.
    It is meant to be inherited and implemented fully by test classes which use nornir.

    :param _transformed_result: All parsed results from a nornir task
    :param nornir: The initialized nornir instance

    """

    #: The path to a nornir configuration file.
    #: https://nornir.readthedocs.io/en/stable/configuration/index.html
    DEFAULT_NORNIR_CONFIG_FILE = "nr-config.yaml"

    def __init__(
        self, nuts_parameters: Any = None, pytestconfig: Optional[Config] = None
    ):
        super().__init__(nuts_parameters, pytestconfig)
        self.nornir: Optional[Nornir] = None

    def initialize(self) -> None:
        if self.pytestconfig:
            config_file = pathlib.Path(
                self.pytestconfig.getoption("nornir_configuration")
            )
        else:
            config_file = pathlib.Path(self.DEFAULT_NORNIR_CONFIG_FILE)

        self.nornir = InitNornir(
            config_file=str(config_file),
            logging={"enabled": False},
        )

    def nuts_task(self) -> Callable[..., Result]:
        """
        Returns the task that nornir should execute for the test module.

        :return: A task as defined by one of nornir's plugins
            or a function that calls a nornir task
        """
        raise NotImplementedError

    def nornir_filter(self) -> F:
        """
        :return: A nornir filter that is applied to the nornir instance
        """
        return filter_hosts(self.nuts_parameters["test_data"])

    def general_result(self) -> AggregatedResult:
        """
        Nornir is run with the defined task, additional arguments,
            a nornir filter and returns the raw result from nornir.
        If the setup/teardown methods are overwritten, these are executed as well.

        :return: The raw result as provided by nornir's executed task
        """
        if not self.nornir:
            raise NutsSetupError("Nornir instance not found in context object")
        self.setup()
        nornir_filter = self.nornir_filter()

        if nornir_filter:
            selected_hosts = self.nornir.filter(nornir_filter)

        else:
            selected_hosts = self.nornir

        if not selected_hosts.inventory.hosts:
            if nornir_filter:
                raise NutsSetupError(
                    f'Host(s) "{",".join(nornir_filter.filters["name__any"])}" '
                    f"not found in the inventory."
                )
            else:
                raise NutsSetupError("No Hosts found, is the nornir inventory empty?")

        overall_results = selected_hosts.run(
            task=self.nuts_task(), **self.nuts_arguments()
        )

        self.teardown()
        return overall_results

    def setup(self) -> None:
        """
        Defines code which is executed before the nornir task.
        """
        pass

    def teardown(self) -> None:
        """
        Defines code which is executed after the nornir task.
        """
        pass



from typing import Dict, Optional


class InterfaceNameConverter:
    def __init__(self, conversion_rules: Optional[Dict[str, str]] = None):
        self.interface_conversion_rules = conversion_rules or {
            "GigabitEthernet": "Gi",
            "FastEthernet": "Fa",
        }

    def shorten_interface_name(self, name: str) -> str:
        for key, value in self.interface_conversion_rules.items():
            if name.lower().startswith(key.lower()):
                return value + name[len(key) :]  # noqa: E203
        return name

    def expand_interface_name(self, name: str) -> str:
        for key, value in self.interface_conversion_rules.items():
            if name.lower().startswith(value.lower()):
                return key + name[len(value) :]  # noqa: E203
        return name



"""Query interfaces and their information of a device."""
from typing import Dict, Callable, List, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.result import AbstractHostResultExtractor, NutsResult


class InterfacesExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[str, Dict[str, Any]]:
        return self._simple_extract(single_result)["interfaces"]


class InterfacesContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["interfaces"]}

    def nuts_extractor(self) -> InterfacesExtractor:
        return InterfacesExtractor(self)


CONTEXT = InterfacesContext


class TestNapalmInterfaces:
    @pytest.mark.nuts("name,is_enabled")
    def test_is_enabled(
        self, single_result: NutsResult, name: str, is_enabled: bool
    ) -> None:
        assert single_result.result[name]["is_enabled"] == is_enabled

    @pytest.mark.nuts("name,is_up")
    def test_is_up(self, single_result: NutsResult, name: str, is_up: bool) -> None:
        assert single_result.result[name]["is_up"] == is_up

    @pytest.mark.nuts("name,mac_address")
    def test_mac_address(
        self, single_result: NutsResult, name: str, mac_address: str
    ) -> None:
        assert single_result.result[name]["mac_address"] == mac_address

    @pytest.mark.nuts("name,mtu")
    def test_mtu(self, single_result: NutsResult, name: str, mtu: int) -> None:
        assert single_result.result[name]["mtu"] == mtu

    @pytest.mark.nuts("name,speed")
    def test_speed(self, single_result: NutsResult, name: str, speed: int) -> None:
        assert single_result.result[name]["speed"] == speed



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nuts', 'nuts.base_tests', 'nuts.helpers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'nornir-napalm>=0.1.1,<0.2.0',
 'nornir-netmiko>=0.1.1,<0.2.0',
 'nornir-utils>=0.1.0,<0.2.0',
 'nornir>=3.0.0,<4.0.0',
 'pytest>=6.1.1,<7.0.0']

entry_points = \
{'pytest11': ['nuts = nuts.plugin']}

setup_kwargs = {
    'name': 'nuts',
    'version': '3.1.0',
    'description': 'Network Unit Testing System Component of NetTowel',
    'long_description': '# NetTowel Network Unit Testing System\n\n## Introduction\n\nThe NetTowel Network Unit Testing System or "nuts" in short is the testing component of the NetTowel Project, which is developed at the Institute of Networked Solutions in Rapperswil, Switzerland.\nIt draws on the concept of unit tests, known from the domain of programming, and applies it to the domain of networking.\n\nOne major difference between unit tests in programming and \nnetwork tests is the definition of what a test actually is. \nIn programming, unit tests normally focus on testing edge cases, \nsince the amount of non-edge cases is not definable.\nIn the network testing domain, tests are less about edge cases, but more about testing existing network states with \npre-defined test cases. Such a single test case might be "can host A reach neighbors X, Y, Z?" on many different devices. \nThis is what nuts tries to achieve:\nApply test cases based on your pre-defined network topology to your actual network and have the tests confirm the correct state.\n\n## Installation Instructions\n\n### Using pip\n\nRun `pip install nuts` \n\n### Using poetry\n\nNuts uses [poetry](https://python-poetry.org/) as a dependency manager.\n\n1. [Install poetry](https://python-poetry.org/docs/#installation).\n2. Clone this repository.\n3. Run `$ poetry install`\n\n## How It Works: Test Bundles and Test Definitions\n\nThe project relies on the [pytest framework](https://docs.pytest.org/) to setup and execute the tests. \nNuts itself is written as a custom pytest plugin. In the background, [nornir](https://nornir.readthedocs.io/) \nexecutes specific network tasks for the actual tests.\n\nNuts treats the test definition and the so-called test bundle as separate entities. The *test definition* is modeled as a custom `pytest.Class`, and a predefined set of test definitions can be found in the nuts module `base_tests`. New test definitions can be added easily by the user of the plugin.\n\nThe *test bundle* is a file that is parsed by pytest. The file provides data on the desired network state and describes which test definitions should be collected and executed by pytest. \nThe structure of the test bundle should enable people without in-depth python knowledge to add new test bundles or update existing ones to reflect changes in the network. \n\nWhile the readme here is only a short overview, find the [documentation of nuts on readthedocs](https://nuts.readthedocs.io/en/latest/).\n\n### Test Bundle Structure\n\nCurrently only yaml files are supported as test bundles, \nbut other sources such as other file formats or database entries can be considered in later nuts versions.\n\nEach test bundle contains the following structure:\n```yaml\n---\n- test_module: <module that contains the test class> # optional\n  test_class: <name of the test class>\n  label: <label to uniquely identify the test> # optional \n  test_execution: <additional data used to execute the test> # optional\n  test_data: <data used to generate the test cases>\n...\n```\n`test_module`: The full path of the python module that contains the test class to be used.\nThis value is optional if the test class is registered in `index.py` of the pytest-nuts plugin.\nNote that it can be relevant in which directory `pytest` is started if local test modules are used. Using `test_modules` allows you to write your own test classes. **Note: We currently do not support self-written test modules, since upcoming refactorings might introduce breaking changes.**\n\n`test_class`: The name of the python class which contains the tests that should be executed.\nNote that currently every test in this class will be executed.\n\n`label`: Additional identifier that can be used to distinguish between multiple occurrences of the same \n test class in a test bundle.\n\n`test_execution`: Data that is exposed as part of the `nuts_parameters` property. \nBy convention this contains additional information that is passed directly to the nornir task in the background. \nTherefore the key-value pairs must be consistent with the key-value pairs of the specific nornir task. \nAs an example, the test definition `napalm_ping.py` calls a nornir task to execute napalm\'s ping-command. \nThis allows the additional `max_drop` parameter in `test_execution`, since it is in turn pre-defined by napalm.\n\n`test_data`: Data that is used to parametrize the tests in the test class which have the `pytest.mark.nuts` annotation. It is additionally part of the `nuts_parameters` property.\n\n### Example: CDP Neighbors\nExample of a test bundle for `TestNetmikoCdpNeighbors` which tests that `R1` is a CDP Neighbor of both `R2` and `R3`.\nThis example creates three different tests, one for each entry in the `test_data` list.\n\n```yaml\n---\n- test_module: nuts.base_tests.netmiko_cdp_neighbors\n  test_class: TestNetmikoCdpNeighbors\n  test_data:\n    - host: R1\n      local_port: GigabitEthernet3\n      destination_host: R2\n      management_ip: 172.16.12.2\n      remote_port: GigabitEthernet2\n    - host: R1\n      local_port: GigabitEthernet4\n      destination_host: R3\n      management_ip: 172.16.13.3\n      remote_port: GigabitEthernet2\n    - host: R2\n      local_port: GigabitEthernet2\n      destination_host: R1\n      management_ip: 172.16.12.1\n      remote_port: GigabitEthernet3\n...\n```\n\n### How the Test Bundle Is Converted to a Pytest Test\n\nWhen nuts is executed, pytest converts the test bundles (the yaml files) into tests. During test collection, the custom pytest marker `nuts` uses the data that has been defined in the test bundle mentioned above. \nThis annotation is a wrapper around the `pytest.mark.parametrize` annotation and allows the plugin to use the data entries from the test bundle. For each entry in the `test_data` section of the test bundle, the custom marker generates a single test case. To achieve this, the plugin transforms the entries into n-tuples, since `pytest.mark.parametrize` expects a list of n-tuples as input. \n\nThe custom nuts marker takes two arguments: The first argument of the annotation determines the required fields. \nFor each entry in `test_data` these fields are extracted and transformed to a tuple considering the correct order.\nIf any of these fields are not present in an entry of `test_data`, the corresponding test case will be skipped.\nA second argument determines optional fields that can also be used in a test case as well - non-present values are passed into the function as `None`.\n\nThe following test-run of CDP neighbors for example checks the local port:\n\n```python\nclass TestNetmikoCdpNeighbors:       \n    @pytest.mark.nuts("remote_host,local_port")\n    def test_local_port(self, single_result, remote_host, local_port):\n        assert single_result.result[remote_host]["local_port"] == local_port        \n```\n\n\nThe required fields are `host`, `remote_host` and `local_port` - they must be present in the custom marker, \nbut also be provided as argument to the test method itself.\n\n`single_result` uses the `host` field and provides the result that has been processed via the specific context of a test.\n\n### Test classes and their context\nEach test module implements a context class to provide module-specific functionality to its tests. This context class is a  `NutsContext` or a subclass of it. \nThis guarantees a consistent interface across all tests for test setup and execution. \nCurrently, the predefined test classes use [nornir](https://nornir.readthedocs.io/en/latest/) in order to communicate \nwith the network devices. Those test classes derive all from a more specific `NornirNutsContext`, \nwhich provides a nornir instance and nornir-specific helpers. In the example above, it is a class called `CdpNeighborsContext` that derives from `NornirNutsContext`.\n\nIf you want to learn more how nuts works but do not have a running network in the background, there\'s a nuts showcase - an offline test class that displays the basic functionality of nuts. See the [tutorial](https://nuts.readthedocs.io/en/latest/tutorial/firststeps.html) for further information.\n\n## Develop Your Own Test Classes\n\nNuts is essentially designed as a pytest-plugin and it is possible to add your own, self-written test classes. \nA dev documentation on how to write your own test classes is planned for a future release.\nStill, it is possible to write your own test classes nevertheless, even if we cannot guarantee that upcoming planned refactorings  do not introduce breaking changes.\n\n## Community-provided test classes\n\n* [Cisco IOSXE Test Class](https://github.com/briantsaunders/nuts-cisco-iosxe-tests) by [briantsaunders](https://github.com/briantsaunders)\n\n# Thanks\n\n* [Matthias Gabriel](https://github.com/MatthiasGabriel), who laid the foundations of nuts.\n* [Florian Bruhin (The Compiler)](https://github.com/The-Compiler) for invaluable feedback and advice.\n',
    'author': 'Lukas Murer, Méline Sieber, Urs Baumann, Matthias Gabriel, Florian Bruhin',
    'author_email': None,
    'maintainer': 'Lukas Murer',
    'maintainer_email': 'lukas.murer@ost.ch',
    'url': 'https://github.com/INSRapperswil/Nuts',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Converts a test bundle (YAML file) into a test class for pytest.
Based on https://docs.pytest.org/en/stable/example/nonpython.html#yaml-plugin
"""
import importlib
import types
from importlib import util
from typing import Iterable, Union, Any, Optional, List, Set, Dict, Tuple

import pytest
import yaml
from _pytest import nodes
from _pytest.mark import ParameterSet
from _pytest.nodes import Node
from _pytest.python import Metafunc

from nuts.helpers.errors import NutsUsageError, NutsSetupError
from nuts import index


class NutsYamlFile(pytest.File):
    """
    Collect tests from a yaml file.
    """

    def collect(self) -> Iterable[Union[nodes.Item, nodes.Collector]]:
        # path uses pathlib.Path and is meant to replace fspath, which uses
        # py.path.local. Both variants will be used for some time in parallel within
        # pytest. If fspath is used in a newer pytest version, it triggers a deprecation
        # warning. We therefore use a wrapper that can use both path types
        if hasattr(self, "path"):
            yield from self._collect_path()
        else:
            yield from self._collect_fspath()

    def _collect_path(self) -> Iterable[Union[nodes.Item, nodes.Collector]]:
        try:
            with self.path.open() as fo:  # type: ignore[attr-defined]
                raw = yaml.safe_load(fo)
        except OSError as ex:
            raise NutsSetupError(
                f"Could not open YAML file containing test bundle:\n{ex}"
            )

        for test_entry in raw:
            module = find_and_load_module(test_entry)
            yield NutsTestFile.from_parent(
                self,
                path=self.path,  # type: ignore[attr-defined, call-arg]
                obj=module,
                test_entry=test_entry,
            )

    def _collect_fspath(self) -> Iterable[Union[nodes.Item, nodes.Collector]]:
        try:
            with self.fspath.open() as f:  # type: ignore[union-attr]
                raw = yaml.safe_load(f)
        except OSError as e:
            raise NutsSetupError(
                f"Could not open YAML file containing test bundle:\n{e}"
            )

        for test_entry in raw:
            module = find_and_load_module(test_entry)
            yield NutsTestFile.from_parent(
                self, fspath=self.fspath, obj=module, test_entry=test_entry
            )


def find_and_load_module(test_entry: Dict[str, str]) -> types.ModuleType:
    test_class = test_entry.get("test_class")
    if not test_class:
        raise NutsUsageError("Class name of the specific test missing in YAML file.")
    module_path = find_module_path(test_entry.get("test_module"), test_class)
    return load_module(module_path)


def find_module_path(module_path: Optional[str], class_name: str) -> str:
    if not module_path:
        module_path = index.find_test_module_of_class(class_name)
        if not module_path:
            raise NutsUsageError(
                "A module that corresponds to the test_class "
                f"called {class_name} could not be found. "
                f"Did you forget to put it into the index (nuts/index.py)?"
            )
    return module_path


def load_module(module_path: str) -> types.ModuleType:
    spec = util.find_spec(module_path)
    if spec is None:
        raise NutsUsageError(
            f"Module path called {module_path} not found. "
            f"Did you spell the module and class name correctly?"
        )
    module = util.module_from_spec(spec)
    # https://github.com/python/typeshed/issues/2793
    assert isinstance(spec.loader, importlib.abc.Loader)
    spec.loader.exec_module(module)
    return module


class NutsTestFile(pytest.Module):
    """
    Custom nuts collector for test classes and functions.
    """

    def __init__(self, obj: Any, test_entry: Any, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.obj = obj
        self.test_entry = test_entry

    def collect(self) -> Iterable[Union[nodes.Item, nodes.Collector]]:
        """
        Collects a single NutsTestClass instance from this NutsTestFile.
        At the start inject setup_module fixture and parse all fixtures from the module.
        This is directly adopted from pytest.Module.
        """

        self._inject_setup_module_fixture()
        self._inject_setup_function_fixture()
        self.session._fixturemanager.parsefactories(self)

        class_name = self.test_entry["test_class"]
        label = self.test_entry.get("label")
        name = class_name if label is None else f"{class_name} - {label}"

        test_data = self.test_entry.get("test_data", [])
        test_execution = self.test_entry.get("test_execution")
        yield NutsTestClass.from_parent(
            self,
            name=name,
            class_name=class_name,
            test_data=test_data,
            test_execution=test_execution,
        )


class NutsTestClass(pytest.Class):
    """
    Custom nuts test collector for test methods.
    Initialises a corresponding context with externally provided parameters.
    """

    def __init__(self, parent: NutsTestFile, name: str, class_name: str, **kw: Any):
        super().__init__(name, parent=parent)
        self.params: Any = kw
        self.name: str = name
        self.class_name: str = class_name

    def _getobj(self) -> Any:
        """
        Get the underlying Python object.
        Overwritten from PyobjMixin to separate name and classname.
        This allows to group multiple tests of the same class with
        different parameters to be grouped separately.
        """
        # cf. https://github.com/pytest-dev/pytest/blob/master/src/_pytest/python.py
        assert self.parent is not None
        obj = self.parent.obj  # type: ignore[attr-defined]
        return getattr(obj, self.class_name)

    @classmethod
    def from_parent(  # type: ignore[override]
        cls, parent: Node, *, name: str, obj: Any = None, **kw: Any
    ) -> Any:
        """The public constructor."""
        # mypy throws an error because the parent class (pytest.Class) does not accept
        # additional **kw.
        # This has been fixed in: https://github.com/pytest-dev/pytest/pull/8367
        # and will be part of a future pytest release. Until then, mypy is instructed
        # to ignore this error
        return cls._create(parent=parent, name=name, obj=obj, **kw)


def get_parametrize_data(
    metafunc: Metafunc,
    fields_str: Optional[str] = None,
    optional_fields_str: Optional[str] = None,
) -> Tuple[List[str], List[ParameterSet]]:
    """
    Transforms externally provided parameters to be used in parametrized tests.

    For every single test run one entry from the test_data section in the yaml file is
    injected as a first entry to be parametrized (`nuts_test_entry`).
    In doing so, the `single_result` fixture in `plugin.py` can pass on
    the full entry to the extractor. The extractor can then decide in its
    own `single_result` method which property of the entry should be picked as key
    (e.g. `[host]` or `[host][destination]`).

    :param metafunc: The annotated test function that will use the parametrized data.
    :param fields_str: The fields used in a test, coming from pytest.mark.nuts.
    :param optional_fields_str: Fields which are optional, coming from pytest.mark.nuts.
    :return: A tuple with 2 entries:
       - List of field names.
       - List of tuples that contain each the parameters for a test.
    """
    if fields_str is None:
        fields = []
    else:
        fields = [field.strip() for field in fields_str.split(",")]

    if optional_fields_str is None:
        optional_fields = set()
    else:
        optional_fields = {field.strip() for field in optional_fields_str.split(",")}
    required_fields = set(fields) - optional_fields

    assert metafunc.definition.parent is not None
    nuts_test_instance = metafunc.definition.parent.parent
    data = getattr(nuts_test_instance, "params")
    return (
        ["nuts_test_entry", *fields],
        dict_to_tuple_list(data["test_data"], fields, required_fields),
    )


def dict_to_tuple_list(
    test_data: List[Dict[str, Any]], fields: List[str], required_fields: Set[str]
) -> List[ParameterSet]:
    return [
        wrap_if_needed(entry, required_fields, dict_to_tuple(entry, fields))
        for entry in test_data
    ]


def wrap_if_needed(
    entry: Dict[str, Any],
    required_fields: Set[str],
    present_fields: Tuple[Optional[Any], ...],
) -> ParameterSet:
    missing_fields = required_fields - set(entry)
    if not missing_fields:
        return pytest.param(entry, *present_fields)
    return pytest.param(
        entry,
        *present_fields,
        marks=pytest.mark.skip(
            f"required values {missing_fields} not present in test-bundle"
        ),
    )


def dict_to_tuple(
    source: Dict[str, Any], fields: List[str]
) -> Tuple[Optional[Any], ...]:
    ordered_fields = [source.get(field) for field in fields]
    return tuple(ordered_fields)



"""Query LLDP neighbors of a device."""
from typing import Dict, Callable, List, Any

import pytest
from nornir.core.task import MultiResult, Result
from nornir_napalm.plugins.tasks import napalm_get

from nuts.context import NornirNutsContext
from nuts.helpers.converters import InterfaceNameConverter
from nuts.helpers.result import AbstractHostResultExtractor, NutsResult


class LldpNeighborsExtractor(AbstractHostResultExtractor):
    def single_transform(self, single_result: MultiResult) -> Dict[str, Dict[str, Any]]:
        neighbors = self._simple_extract(single_result)["lldp_neighbors_detail"]
        return {
            peer: self._add_custom_fields(details[0])
            for peer, details in neighbors.items()
        }

    def _add_custom_fields(self, element: Dict[str, Any]) -> Dict[str, Any]:
        element = self._add_expanded_remote_port(element)
        element = self._add_remote_host(element)
        return element

    def _add_remote_host(self, element: Dict[str, Any]) -> Dict[str, Any]:
        element["remote_host"] = element["remote_system_name"]
        return element

    def _add_expanded_remote_port(self, element: Dict[str, Any]) -> Dict[str, Any]:
        element[
            "remote_port_expanded"
        ] = InterfaceNameConverter().expand_interface_name(element["remote_port"])
        return element


class LldpNeighborsContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return napalm_get

    def nuts_arguments(self) -> Dict[str, List[str]]:
        return {"getters": ["lldp_neighbors_detail"]}

    def nuts_extractor(self) -> LldpNeighborsExtractor:
        return LldpNeighborsExtractor(self)


CONTEXT = LldpNeighborsContext


class TestNapalmLldpNeighbors:
    @pytest.mark.nuts("local_port,remote_host")
    def test_remote_host(
        self, single_result: NutsResult, local_port: str, remote_host: str
    ) -> None:
        lldp_neighbor_entry = single_result.result[local_port]
        assert lldp_neighbor_entry["remote_host"] == remote_host

    @pytest.mark.nuts("local_port,remote_port")
    def test_remote_port(
        self, single_result: NutsResult, local_port: str, remote_port: str
    ) -> None:
        lldp_neighbor_entry = single_result.result[local_port]
        assert (
            lldp_neighbor_entry["remote_port"] == remote_port
            or lldp_neighbor_entry["remote_port_expanded"] == remote_port
        )


class TestNapalmLldpNeighborsCount:
    @pytest.mark.nuts("neighbor_count")
    def test_neighbor_count(self, single_result, neighbor_count):
        assert neighbor_count == len(single_result.result)



"""Let a device ping another device."""
from enum import Enum
from typing import Dict, Callable, Any, List

import pytest
from nornir.core import Task
from nornir.core.task import Result
from nornir_napalm.plugins.tasks import napalm_ping

from nuts.context import NornirNutsContext
from nuts.helpers.result import (
    NutsResult,
    AbstractHostDestResultExtractor,
)


class Ping(Enum):
    FAIL = 0
    SUCCESS = 1
    FLAPPING = 2


class PingExtractor(AbstractHostDestResultExtractor):
    def single_transform(self, single_result: Result) -> Ping:
        assert single_result.host is not None
        assert single_result.result is not None
        max_drop = self._allowed_max_drop_for_destination(
            single_result.host.name,
            single_result.destination,  # type: ignore[attr-defined] # see below
        )
        return self._map_result_to_enum(single_result.result, max_drop)

    def _allowed_max_drop_for_destination(self, host: str, dest: str) -> int:
        """
        Matches the host-destination pair from a nornir task to the
        host-destination pair from test_data and retrieves its
        max_drop value that has been defined for that pair.

        :param host: host entry from the nornir task
        :param dest: destination that was pinged by a host from the nornir task
        :return: max_drop value from test_data
        """
        test_data: List[Dict[str, Any]] = self._nuts_ctx.nuts_parameters["test_data"]
        for entry in test_data:
            if entry["host"] == host and entry["destination"] == dest:
                return entry["max_drop"]
        return 0

    def _map_result_to_enum(
        self, result: Dict[str, Dict[Any, Any]], max_drop: int
    ) -> Ping:
        """
        Evaluates the ping that has been conducted with nornir and matches it
        to a Ping-Enum which can be either FAIL, SUCCESS or FLAPPING.

        FAIL: Packet loss equals probes sent.
        SUCCESS: Packet loss is below or equal max_drop.
        FLAPPING: Everything else.

        :param result: a single nornir Result
        :param max_drop: max_drop threshold
        :return: evaluated ping result
        """
        if result["success"]["packet_loss"] == result["success"]["probes_sent"]:
            return Ping.FAIL
        if result["success"]["packet_loss"] <= max_drop:
            return Ping.SUCCESS
        else:
            return Ping.FLAPPING

    def single_result(self, nuts_test_entry: Dict[str, Any]) -> NutsResult:
        host = nuts_test_entry["host"]
        destination = nuts_test_entry["destination"]
        assert (
            host in self.transformed_result
        ), f"Host {host} not found in aggregated result."
        assert (
            destination in self.transformed_result[host]
        ), f"Destination {destination} not found in result."
        return self.transformed_result[host][destination]


class PingContext(NornirNutsContext):
    def nuts_task(self) -> Callable[..., Result]:
        return self.napalm_ping_multi_dests

    def napalm_ping_multi_dests(self, task: Task, **kwargs: Any) -> Result:
        """
        One host pings all destinations as defined in the test bundle.

        Note: The destination is not included in the nornir result if the ping fails.
        Therefore we cannot know which destination was not reachable,
        so we must patch the destination onto the result object to know later which
        host-destination pair actually failed.

        :param task: nornir task for ping
        :param kwargs: arguments from the test bundle for the napalm ping task, such as
        count, ttl, timeout
        :return: all pinged destinations per host
        """
        destinations_per_hosts = [
            entry["destination"]
            for entry in self.nuts_parameters["test_data"]
            if entry["host"] == task.host.name
        ]
        for destination in destinations_per_hosts:
            result = task.run(task=napalm_ping, dest=destination, **kwargs)
            result[0].destination = destination  # type: ignore[attr-defined]
        return Result(host=task.host, result="All pings executed")

    def nuts_extractor(self) -> PingExtractor:
        return PingExtractor(self)


CONTEXT = PingContext


class TestNapalmPing:
    @pytest.mark.nuts("expected")
    def test_ping(self, single_result: NutsResult, expected: str) -> None:
        assert single_result.result.name == expected
